#ifndef __TRANSLATIONRULE_H
#define __TRANSLATIONRULE_H

class DoubleDatabase;
class Database;
class ItemSet;
class ItemSetCollection;

#include <Bass.h>
#include <itemstructs/ItemSet.h>

enum RuleDirection {
	RuleBoth,
	RuleToRight,
	RuleToLeft,
};

class TranslationRule;
typedef std::list<TranslationRule*> rulelist;

class TranslationRule {
public:
	TranslationRule(ItemSet *setL, const RuleDirection dir, ItemSet *setR);
	TranslationRule(ItemSet *setL, const RuleDirection dir, ItemSet *setR, double encodedSize);
	TranslationRule(const TranslationRule &tt);
	virtual ~TranslationRule();
	virtual TranslationRule* Clone();

	// Get/set
	ItemSet*		GetLeft()			{ return mLeftSet; }
	ItemSet*		GetRight()			{ return mRightSet; }
	RuleDirection	GetDirection()		{ return mDirection; }
	void			SetDirection(const RuleDirection dir);
	double			GetEncodedSize()	{ return mEncodedSize; }

	// Serialization
	string			ToString();

protected:
	// Helpers
	void			ComputeEncodedSize(); // rule should always keep encSize up to date

	// Members
	ItemSet*		mLeftSet;
	RuleDirection	mDirection;
	ItemSet*		mRightSet;

	double			mEncodedSize;
};

#endif // __TRANSLATIONRULE_H
